import torch
from torch import nn, Tensor
from zeta.nn import MambaBlock, OutputHead
from zeta.nn.attention.multiquery_attention import MultiQueryAttention

class MambaFormerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        d_state: int,
        d_conv: int,
        heads: int,
        dim_head: int,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.dim = dim
        self.d_state = d_state
        self.d_conv = d_conv
        self.heads = heads
        self.dim_head = dim_head
        
        # MambaBlock
        self.input_mamba = MambaBlock(
            dim,
            1,
            d_state,
            d_conv,
            *args,
            **kwargs
        )
        
        # MultiQueryAttention
        self.attn = MultiQueryAttention(
            dim,
            heads,
            *args,
            **kwargs   
        )
    
    def forward(self, x: Tensor):
        # First step is Attention
        skip = x
        
        # Attention
        attended, _, _ = self.attn(x)
        
        # Add residual
        x = attended + skip
        
        # Second step is MambaBlock
        skip_two = x
        
        # Mamba
        x = self.input_mamba(x)
        
        # Add residual and return
        return x + skip_two
    

class MambaFormer(nn.Module):
    def __init__(
        self,
        dim: int,
        num_tokens: int,
        depth: int,
        d_state: int,
        d_conv: int,
        heads: int,
        dim_head: int,
        return_tokens: bool = True,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.dim = dim
        self.num_tokens = num_tokens
        self.depth = depth
        self.d_state = d_state
        self.d_conv = d_conv
        self.heads = heads
        self.dim_head = dim_head
        self.return_tokens = return_tokens
        
        # Embedding
        self.embed = nn.Embedding(num_tokens, dim)
        
        # Mamba as input
        self.input_mamba = MambaBlock(
            dim,
            1,
            d_state,
            d_conv,
            *args,
            **kwargs
        )
        
        # Layers of Mamba Former Block
        self.layers = nn.ModuleList([
            MambaFormerBlock(
                dim,
                d_state,
                d_conv,
                heads,
                dim_head,
            )
            for _ in range(depth)
        ])
        
        
        # Norm
        self.norm = nn.LayerNorm(dim)
        
    def forward(self, x):
        # Embed tokens to get tensors
        x = self.embed(x)
        
        # Normalize
        x = self.norm(x)
        
        # Then skip
        skip = x
        
        # Mamba
        x = self.input_mamba(x) + skip
        
        # MambaFormer Blocks
        for layer in self.layers:
            x = layer(x) + x
        
        if self.return_tokens:
            out = OutputHead(self.dim, -1)(x)
            return out
        else: 
            return x
        



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mamba_former']

package_data = \
{'': ['*']}

install_requires = \
['einops', 'swarms', 'torch', 'zetascale']

setup_kwargs = {
    'name': 'mamba-former',
    'version': '0.0.1',
    'description': 'Paper - Pytorch',
    'long_description': '[![Multi-Modality](agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n# MambaFormer\nImplementation of MambaFormer in Pytorch ++ Zeta from the paper: "Can Mamba Learn How to Learn? A Comparative Study on In-Context Learning Tasks"\n\n## install\n`pip3 install mamba-former`\n\n## usage\n```python\nimport torch \nfrom mamba_former.main import MambaFormer\n\n# Forward pass example\nx = torch.randint(1, 1000, (1, 100)) # Token\n# Tokens are integrers\n\n# Model\nmodel = MambaFormer(\n    dim = 512,\n    num_tokens = 1000,\n    depth = 6,\n    d_state = 512,\n    d_conv = 128,\n    heads = 8,\n    dim_head = 64,\n    return_tokens = True\n)\n\n# Forward\nout = model(x)\nprint(out)\nprint(out.shape)\n```\n\n\n# License\nMIT\n',
    'author': 'Kye Gomez',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/MambaFormer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
